import pymysql
from ..base import Base
from ..dialect import Dialect

class DialectMysql(Dialect):
	pass

class Mysql(Base):

	def __init__(self, *args, **kwargs):
		super(Mysql, self).__init__(*args, **kwargs)
		self.cursor = self.connection.cursor()
		self.dialect(DialectMysql)
	
	def _connect(self, *args, **kwargs):
		arguments = {
			'host': 'localhost',
			'port': 3306
		}
		arguments.update(kwargs)
		return pymysql.connect(**arguments)
		


from collections import OrderedDict
from .builder import Builder
from .record import Records
from .dialect import Dialect

class Base(object):
	
	def __init__(self, *args, **kwargs):
		if 'logging' in kwargs:
			self.logging = kwargs['logging']
			del kwargs['logging']
		else:
			self.logging = False

		self._dialect = None
		if 'dialect' in kwargs:
			self._dialect = kwargs['dialect']
			self.dialect(kwargs['dialect'])
			del kwargs['dialect']
	
		# in case self._connect raises error
		self.connection = None
		self.connection = self._connect(*args, **kwargs)
		self.cursor     = self.connection.cursor()
		self.history    = []
		self.errors     = []
		self.sql        = None

	@property
	def builder(self):
		return Builder(dialect = self._dialect)
		
	def id (self):
		return self.cursor.lastrowid
		
	def _connect(self, *args, **kwargs):
		raise NotImplementedError('API not implemented.')
		
	def close(self):
		self.connection.close()

	def dialect(self, dial = None):
		dial = dial or Dialect
		self._dialect   = dial
		Builder.DIALECT = dial
		
	def __del__(self):
		if self.connection:
			self.close()
			
	def last(self):
		return self.history[-1] if self.history else ''
		
	def log(self):
		return self.history
		
	def error(self):
		return self.errors
		
	def commit(self):
		try:
			self.connection.commit()
		except Exception as ex:
			self.connection.rollback()
			raise ex
	
	# If data is an ordered dict, then datas could be tuples
	# otherwise, datas also should be dicts
	def insert(self, table, fields, *values, **kwargs):
		sql = self.builder.insert(table, fields, *values)
		return self.query(sql, kwargs.get('commit', True))
			
	def update(self, table, data, where = None, commit = True):
		sql = self.builder.update(table, data, where)
		return self.query(sql, commit = commit)
	
	# where required to avoid all data deletion
	def delete(self, table, where, commit = True):
		sql = self.builder.delete(table, where)
		return self.query(sql, commit = commit)		
		
	def select(self, table, columns = '*', where = None, join = None, distinct = False, newtable = None, sub = None, commit = False):
		sql = self.builder.select(table, columns, where, join, distinct, newtable, sub)
		return self.query(sql, commit)

	def union(self, *queries, **kwargs):
		sql = self.builder.union(*queries)
		return self.query(sql, commit = kwargs.get('commit', False))
			
	def has(self, table, where = None, join = None, distinct = False):
		rs = self.select(table, '*', where, join)
		return bool(rs.first())
		
	def get(self, table, columns = '*', where = None, join = None):
		rs = self.select(table, columns, where, join)
		return rs.first()[0]
			
	def query(self, sql, commit = True):
		self.sql = str(sql).strip()
		if self.logging:
			self.history.append(self.sql)
		else:
			self.history = [self.sql]
		try:
			self.cursor.execute(self.sql)
			if commit:
				self.commit()
			if self.sql.upper().startswith('SELECT'):
				return Records(self.cursor)
			else:
				return True
		except Exception as ex:
			from sys import stderr
			self.errors.append(ex)
			stderr.write(self.sql + '\n')
			raise
			



class UpdateParseError(Exception):
	pass
	
class FieldParseError(Exception):
	pass
	
class TableParseError(Exception):
	pass
	
class WhereParseError(Exception):
	pass

class JoinParseError(Exception):
	pass

class AnyAllSomeParseError(Exception):
	pass

class LimitParseError(Exception):
	pass

class InsertParseError(Exception):
	pass

class RecordKeyError(KeyError):
	pass

class RecordAttributeError(AttributeError):
	pass



VERSION = '0.0.1'

class utils(object):

	@staticmethod
	def alwaysList(x):
		"""
		Always return a list
		"""
		from six import string_types
		return [y.strip() for y in x.split(',')] if isinstance(x, string_types) else list(x)

	@staticmethod
	def reduce_datetimes(row):
		"""
		Receives a row, converts datetimes to strings.
		"""
		row = list(row)

		for i in range(len(row)):
			if hasattr(row[i], 'isoformat'):
				row[i] = row[i].isoformat()
		return tuple(row)

import importlib
from .builder import Raw, Table, Field
from .dialect import Dialect

DATABASE_TYPES = {
	'Sqlite': ['sqlite', 'sqlite3'],
	'Mysql' : 'mysql',
	'Mssql' : 'mssql',
	'Pgsql' : ['pgsql', 'postgres', 'postgresql'],
}

class Medoo(object):

	def __new__(klass, dbtype, *args, **kwargs):

		for key, val in DATABASE_TYPES.items():
			if not isinstance(val, list):
				val = [val]
			if not dbtype.lower() in val:
				continue

			mod   = importlib.import_module('.database.{}'.format(key.lower()), package = 'medoo')
			klass = getattr(mod, key)
			return klass(*args, **kwargs)

		raise ValueError('Database type not supported: {}.'.format(dbtype))





import six
from .exception import WhereParseError, AnyAllSomeParseError
#import builder

class Dialect(object):

	OPERATOR_MAP = {
		'='  : 'eq',
		'~'  : 'like',
		'!'  : 'ne',
		'!=' : 'ne',
		'<>' : 'between',
		'==' : 'is_',
		'is' : 'is_',
	}

	UPDATE_MAP = {
		'=': 'up_eq'
	}

	JOIN_MAP = {
		'>' : 'LEFT JOIN',
		'<' : 'RIGHT JOIN',
		'<>': 'FULL OUTER JOIN',
		'><': 'INNER JOIN'
	}

	@staticmethod
	def quote(item):
		if isinstance(item, six.string_types):
			if item == '*': return item
			return '"%s"' % item.replace('"', '""')
		else:
			return str(item)

	@staticmethod
	def value(item):
		if isinstance(item, six.string_types):
			return "'{}'".format(str(item).replace("'", "''"))
		else:
			return str(item)

	@classmethod
	def limit(klass, limit, offset = None):
		fmt = 'LIMIT {limit}'
		if offset:
			fmt += ' OFFSET {offset}'
		return fmt.format(limit = limit, offset = offset)

	@classmethod
	def up_eq(klass, field, value):
		return '{}={}'.format(field, klass.value(value))

	@classmethod
	def is_(klass, field, value):
		if not value is None:
			raise WhereParseError('IS is only used to tell NULL (None)')
		return '{} IS NULL'.format(field)

	@classmethod
	def eq(klass, field, value):
		from . import builder
		if isinstance(value, (tuple, list)) and len(value) == 1:
			value = value[0]

		if isinstance(value, (tuple, list)):
			return '{} IN ({})'.format(field, ','.join([klass.value(v) for v in value]))
		elif isinstance(value, builder.Builder): # subquery
			return '{} IN ({})'.format(field, klass.value(value))
		else:
			return '{} = {}'.format(field, klass.value(value))

	@classmethod
	def like(klass, field, value):
		if isinstance(value, (tuple, list)) and len(value) == 1:
			value = value[0]

		if isinstance(value, (tuple, list)):
			value = [
				'%{}%'.format(v) if not v.startswith('%') and not v.endswith('%') else v 
				for v in value 
				if isinstance(v, six.string_types)
			]
			return '({})'.format(' OR '.join(['{} LIKE {}'.format(field, klass.value(v)) for v in value]))
		else:
			if isinstance(value, six.string_types) and not value.startswith('%') and not value.endswith('%'):
				value = '%{}%'.format(value)
			return '{} LIKE {}'.format(field, klass.value(value))

	@classmethod
	def ne(klass, field, value):
		from . import builder
		if isinstance(value, (tuple, list)) and len(value) == 1:
			value = value[0]

		if isinstance(value, (tuple, list)):
			return '{} NOT IN ({})'.format(field, ','.join([klass.value(v) for v in value]))
		elif isinstance(value, builder.Builder):
			return '{} NOT IN ({})'.format(field, klass.value(value))
		else:
			return '{} <> {}'.format(field, klass.value(value))

	@classmethod
	def between(klass, field, value):
		if not isinstance(value, (tuple, list)) or len(value) != 2:
			raise WhereParseError('BETWEEN value should a tuple or list with 2 elements.')
		return '{} BETWEEN {} AND {}'.format(field, klass.value(value[0]), klass.value(value[1]))

	@classmethod
	def _default(klass, oprt, field, value):
		from . import builder
		if oprt.lower().endswith('any') or oprt.lower().endswith('all') or oprt.lower().endswith('some'):
			oprt = oprt.upper()
			if not isinstance(value, builder.Term):
				raise AnyAllSomeParseError('Require a subquery for ALL/ANY/SOME statement.')
			value = '({})'.format(value)
		else:
			value = klass.value(value)
		return '{} {} {}'.format(field, oprt, value)

	@classmethod
	def _up_default(klass, oprt, field, value):
		value = klass.value(value)
		return '{}={}{}{}'.format(field, field, oprt, value)

	@classmethod
	def _join_default(klass, jointype):
		return jointype

	@classmethod
	def _operator(klass, oprt, field, value):
		oprt = oprt or '='
		oprtmap = klass.OPERATOR_MAP
		if oprt in oprtmap:
			oprt = oprtmap[oprt]
		else:
			su = super(klass, klass)
			if hasattr(su, 'OPERATOR_MAP'):
				oprt = su.OPERATOR_MAP.get(oprt, oprt)
		
		if hasattr(klass, oprt):
			return getattr(klass, oprt)(field, value)
		else:
			return klass._default(oprt, field, value)

	@classmethod
	def _update(klass, oprt, field, value):
		oprt = oprt or '='
		upmap = klass.UPDATE_MAP
		if oprt in upmap:
			oprt = upmap[oprt]
		else:
			su = super(klass, klass)
			if hasattr(su, 'UPDATE_MAP'):
				oprt = su.UPDATE_MAP.get(oprt, oprt)
		
		if hasattr(klass, oprt):
			return getattr(klass, oprt)(field, value)
		else:
			return klass._up_default(oprt, field, value)

	@classmethod
	def _join(klass, jointype = None):
		jointype = jointype or '><'
		joinmap  = klass.JOIN_MAP
		if jointype in joinmap:
			jointype = joinmap[jointype]
		else:
			su = super(klass, klass)
			if hasattr(su, 'JOIN_MAP'):
				jointype = su.JOIN_MAP.get(jointype, jointype)
		
		if hasattr(klass, jointype):
			return getattr(klass, jointype)()
		else:
			return klass._join_default(jointype)


